!***********************
! Lorenz 1963 (4DVAR)
! 2012. 8.29 K. Ito
!***********************

  program Lorenz63_4DVAR
  implicit none
  
  !*****
  !---Parameters
  integer, parameter :: nt_asm=1000, nt_prd=1000
  integer, parameter :: obs_interval=100 !interval of time steps
  real(8), parameter :: parm1=10.0d0, parm2=28.0d0, parm3=2.667d0, dt=1.0d-3
  real(8), parameter :: pi=3.14159265358979d0
  !---Physical variable
  real(8) :: x_t(0:nt_asm+nt_prd), y_t(0:nt_asm+nt_prd), z_t(0:nt_asm+nt_prd) 
  real(8) :: x_s(0:nt_asm+nt_prd), y_s(0:nt_asm+nt_prd), z_s(0:nt_asm+nt_prd) 
  real(8) :: x_da(0:nt_asm+nt_prd),y_da(0:nt_asm+nt_prd),z_da(0:nt_asm+nt_prd)
  real(8) :: x_tmp(0:nt_asm),y_tmp(0:nt_asm),z_tmp(0:nt_asm)
  real(8) :: x_b, y_b, z_b ! First guess of initial value
  real(8) :: x_obs(0:nt_asm/obs_interval)
  !---Costfunction and adjoint variable
  real(8) :: J, Jold
  real(8) :: x_save(0:nt_asm), y_save(0:nt_asm), z_save(0:nt_asm)
  real(8) :: adx(0:nt_asm), ady(0:nt_asm), adz(0:nt_asm)
  real(8) :: alpx = 1.0d-4, alpy=1.0d-4, alpz=1.0d-4 ! Coef for minimization
  integer :: iter_max = 5000 ! maximum number of iteration
  real(8) :: cond_iter = 1.0d-4 ! condition for iteration end ((Jold-J)/Jold)
  !---Output control
  integer, parameter :: output_interval=50
  character(len=7) :: obs_chr(0:nt_asm)
  !---Matrix
  real(8) :: B(3,3) !element1:x, element2:y element3:z
  real(8) :: R(1,1) !element1:x, element2:y element3:z
  !---Working variable
  integer :: it
  real(8) :: x_innov
  integer :: iter
  real(8) :: noise1,noise2,gnoise

  !*****Initialization of random number generator
  call random_seed()

  !*****Section 1: Initial value setting
  x_t(0)=10.0d0
  y_t(0)=14.0d0
  z_t(0)=24.0d0
  x_s(0)=11.0d0
  y_s(0)=13.0d0
  z_s(0)=25.0d0
  B(1,1)=1.0d0; B(1,2)=0.0d0; B(1,3)=0.0d0
  B(2,1)=0.0d0; B(2,2)=1.0d0; B(2,3)=0.0d0
  B(3,1)=0.0d0; B(3,2)=0.0d0; B(3,3)=1.0d0
  R(1,1)=0.1d0
  
  !*****Section 2: True Field and Observations
  do it=1,nt_asm+nt_prd
    !---forward time step
    x_t(it) = x_t(it-1)                                                   &
    &       - dt * parm1*(x_t(it-1) - y_t(it-1))
    y_t(it) = y_t(it-1)                                                   &
    &       + dt * (-y_t(it-1) - x_t(it-1) * z_t(it-1) + parm2 * x_t(it-1))
    z_t(it) = z_t(it-1)                                                   &
    &       + dt * (x_t(it-1) * y_t(it-1) - parm3 * z_t(it-1))
    !---Observation
    if ((mod(it,obs_interval) == 0) .and. (it <= nt_asm)) then
      ! Generate Gaussian Noise from uniform random number (Box-Muller)
      call random_number(noise1)
      call random_number(noise2)
      gnoise=sqrt(R(1,1))*sqrt(-2.0d0*log(1.0d0-noise1))*cos(2.0d0*pi*noise2)
      ! Generate observation by adding Gaussian noise to true value
      x_obs(it/obs_interval)=x_t(it) +gnoise
    end if
  end do

  !*****Section 3: Simulation without DA
  do it=1,nt_asm+nt_prd
    x_s(it) = x_s(it-1)                                                   &
    &       - dt * parm1*(x_s(it-1) - y_s(it-1))
    y_s(it) = y_s(it-1)                                                   &
    &       + dt * (-y_s(it-1) - x_s(it-1) * z_s(it-1) + parm2 * x_s(it-1))
    z_s(it) = z_s(it-1)                                                   &
    &       + dt * (x_s(it-1) * y_s(it-1) - parm3 * z_s(it-1))
  end do

  !*****Section 4: Data assimilation: 4DVAR
  x_tmp(0)=x_s(0)
  y_tmp(0)=y_s(0)
  z_tmp(0)=z_s(0)
  x_b=x_tmp(0) !set First guess value
  y_b=y_tmp(0) !set First guess value
  z_b=z_tmp(0) !set First guess value
  do iter=1,iter_max
    !---Section 4-1: initialization for adjoint model run
    J=0.0d0
    adx(0:nt_asm)=0.0d0
    ady(0:nt_asm)=0.0d0
    adz(0:nt_asm)=0.0d0
    !---Section 4-2: Forward model run
    do it=1,nt_asm
      x_tmp(it) = x_tmp(it-1)                                                  &
      &     - dt * parm1*(x_tmp(it-1) - y_tmp(it-1))
      y_tmp(it) = y_tmp(it-1)                                                  &
      &     + dt * (-y_tmp(it-1) - x_tmp(it-1) * z_tmp(it-1) + parm2 * x_tmp(it-1))
      z_tmp(it) = z_tmp(it-1)                                                   &
      &     + dt * (x_tmp(it-1) * y_tmp(it-1) - parm3 * z_tmp(it-1))
      if(mod(it,obs_interval) == 0) then
        !---Calculate costfunction
        J = J + (x_obs(it / obs_interval) - x_tmp(it)) ** 2 / R(1,1)
      end if
    end do
    !---Section 4-3: adjoint model run
    do it=nt_asm,1,-1
      if(mod(it,obs_interval) == 0) then
        !---Calculate misfit and change adjoint variable
        adx(it) = adx(it) + (x_tmp(it)-x_obs(it/obs_interval))/R(1,1)
      end if
      adx(it-1) = adx(it)                            &
      &         - dt * parm1 * adx(it)               &
      &         + dt * (-z_tmp(it) + parm2) * ady(it) &
      &         + dt * y_tmp(it) * adz(it)
      ady(it-1) = ady(it)                 &
      &         + dt * parm1 * adx(it)    &
      &         - dt * ady(it)            &
      &         + dt * x_tmp(it) * adz(it)
      adz(it-1) = adz(it)                 &
      &         - dt * x_tmp(it) * ady(it) &
      &         - dt * parm3 * adz(it)
    end do
    !---Section 4-4: Consider background covariance
    J = J + (x_tmp(0) - x_b)**2 / B(1,1) &
    &     + (y_tmp(0) - y_b)**2 / B(2,2) &
    &     + (z_tmp(0) - z_b)**2 / B(3,3)
    adx(0) = adx(0) + (x_tmp(0) - x_b) / B(1,1)
    ady(0) = ady(0) + (y_tmp(0) - y_b) / B(2,2)
    adz(0) = adz(0) + (z_tmp(0) - z_b) / B(3,3)
    !---Section 4-5: Check the end of iteration
    if((iter > 1) .and. (Jold < J)) then
      x_da(0:nt_asm)=x_tmp(0:nt_asm)
      y_da(0:nt_asm)=y_tmp(0:nt_asm)
      z_da(0:nt_asm)=z_tmp(0:nt_asm)
      write(*,*)
      write(*,*) "Cost function increases from the previous iteration."
      write(*,*) "Replace DA results with those in previous iteration and exit DA."
      write(*,'(A,1X,i3,1X,A,3(F9.3,1X))') &
      &   "iteration=",iter, "; adjoint x,y,z=", adx(0),ady(0),adz(0)
      write(*,'(10X,A,F9.3,1X,A,1X,3(F7.3,2X))') "J=",J, &
      &   "; x(0),y(0),z(0)=>",x_da(0),y_da(0),z_da(0)
      write(*,*)
      exit
    endif
    if((iter > 1) .and. (cond_iter > (Jold-J) / Jold)) then
      x_da(0:nt_asm)=x_tmp(0:nt_asm)
      y_da(0:nt_asm)=y_tmp(0:nt_asm)
      z_da(0:nt_asm)=z_tmp(0:nt_asm)
      write(*,*)
      write(*,*) "Differences between J and Jold become small => exit DA."
      write(*,'(A,1X,i3,1X,A,3(F9.3,1X))') &
      &   "iteration=",iter, "; adjoint x,v=", adx(0),ady(0),adz(0)
      write(*,'(10X,A,F9.3,1X,A,1X,3(F7.3,2X))') "J=",J, &
      &   "; x(0),y(0),z(0)=>",x_da(0),y_da(0),z_da(0)
      write(*,*)
      exit
    endif
    if(iter == iter_max) then
      x_da(0:nt_asm)=x_tmp(0:nt_asm)
      y_da(0:nt_asm)=y_tmp(0:nt_asm)
      z_da(0:nt_asm)=z_tmp(0:nt_asm)
      write(*,*)
      write(*,*) "Iteration maximum reached"
      write(*,'(A,1X,i3,1X,A,3(F9.3,1X))') &
      &   "iteration=",iter, "; adjoint x,v=", adx(0),ady(0),adz(0)
      write(*,'(10X,A,F9.3,1X,A,1X,3(F7.3,2X))') "J=",J, &
      &   "; x(0),y(0),z(0)=>",x_da(0),y_da(0),z_da(0)
      write(*,*)
      exit
    endif
    !---Section 4-6: save values
    Jold=J
    !---Section 4-7: Update of x and v
    x_tmp(0)=x_tmp(0)-alpx*adx(0)
    y_tmp(0)=y_tmp(0)-alpy*ady(0)
    z_tmp(0)=z_tmp(0)-alpz*adz(0)
    !---Print the result of iteration
    write(*,'(A,1X,i3,1X,A,3(F9.3,1X))') &
    &   "iteration=",iter, "; adjoint x,y,z=", adx(0),ady(0),adz(0)
    write(*,'(10X,A,F9.3,1X,A,1X,3(F7.3,2X))') "J=",J, "; x(0),v(0),z(0)=>",x_tmp(0),y_tmp(0),z_tmp(0)
  end do
  
  !*****Section 5: Prediction after Data Assimilation
  do it=nt_asm+1,nt_asm+nt_prd
    x_da(it) = x_da(it-1)                                                   &
    &         - dt * parm1*(x_da(it-1) - y_da(it-1))
    y_da(it) = y_da(it-1)                                                   &
    &     + dt * (-y_da(it-1) - x_da(it-1) * z_da(it-1) + parm2 * x_da(it-1))
    z_da(it) = z_da(it-1)                                                   &
    &       + dt * (x_da(it-1) * y_da(it-1) - parm3 * z_da(it-1))
  end do

  !*****Output on screen
  !---Preparation output: conversion real --> character
  obs_chr(0:nt_asm)="No obs"
  do it=1,nt_asm
    if (mod(it,obs_interval) == 0) then
      write(obs_chr(it), '(F7.3)') x_obs(it/obs_interval)
    end if
  end do
  !---Output Main
  write(*,*)
  write(*,*) "#######################################"
  write(*,*) "###    Identical Twin Experiment    ###"
  write(*,*) "###    Lorenz63: 4DVAR              ###"
  write(*,*) "#######################################"
  write(*,*) 
  write(*,'(A,F7.2,A,F7.2)') "Assmilation Period: t=",        &
  &                          0.0,"-",dt*nt_asm
  write(*,'(A,F7.2,A,F7.2)') "Prediction Period:  t=",         &
  &                          dt*nt_asm,"-", dt*(nt_asm+nt_prd)
  write(*,*) 
  write(*,*) "*******Assimilation Period: x ********"
  write(*,*) " [Time]   [True]  [No Assim]  [Assim]  [Observation]"
  do it=0,nt_asm
    if (mod(it,output_interval) == 0) then
      write(*,'(F7.2,3F10.3,4X,A)') dt*it, x_t(it), x_s(it),x_da(it),obs_chr(it)
    endif
  end do
  write(*,*) "*******Prediction Period: x ********"
  do it=nt_asm+1,nt_asm+nt_prd
    if (mod(it,output_interval) == 0) then
      write(*,'(F7.2,3F10.3)') dt*it, x_t(it), x_s(it),x_da(it)
    endif
  end do
  write(*,*)
  write(*,*) "*******Assimilation Period: y ********"
  write(*,*) " [Time]   [True]  [No Assim]  [Assim]"
  do it=0,nt_asm
    if (mod(it,output_interval) == 0) then
      write(*,'(F7.2,3F10.3)') dt*it, y_t(it), y_s(it), y_da(it)
    endif
  end do
  write(*,*) "*******Prediction Period: y ********"
  do it=nt_asm+1,nt_asm+nt_prd
    if (mod(it,output_interval) == 0) then
      write(*,'(F7.2,3F10.3)') dt*it, y_t(it), y_s(it), y_da(it)
    endif
  end do
  write(*,*)
  write(*,*) "*******Assimilation Period: z ********"
  write(*,*) " [Time]   [True]  [No Assim]  [Assim]"
  do it=0,nt_asm
    if (mod(it,output_interval) == 0) then
      write(*,'(F7.2,3F10.3)') dt*it, z_t(it), z_s(it), z_da(it)
    endif
  end do
  write(*,*) "*******Prediction Period: z ********"
  do it=nt_asm+1,nt_asm+nt_prd
    if (mod(it,output_interval) == 0) then
      write(*,'(F7.2,3F10.3)') dt*it, z_t(it), z_s(it), z_da(it)
    endif
  end do

  end program Lorenz63_4DVAR
