!***********************
! Lorenz 1963 (EnKF)
! 2012. 6. 2 K. Ito
!***********************

  program Lorenz63_EnKF
  implicit none
  
  !*****
  !---Parameters
  integer, parameter :: nt_asm=1000, nt_prd=1000
  integer, parameter :: obs_interval=100
  real(8), parameter :: parm1=10.0d0, parm2=28.0d0, parm3=2.667d0, dt=1.0d-3
  real(8), parameter :: pi=3.14159265358979d0
  integer, parameter :: nens=100
  !---Physical variable
  real(8) :: x_t(0:nt_asm+nt_prd), y_t(0:nt_asm+nt_prd), z_t(0:nt_asm+nt_prd) 
  real(8) :: x_s(0:nt_asm+nt_prd), y_s(0:nt_asm+nt_prd), z_s(0:nt_asm+nt_prd) 
  real(8) :: x_da(0:nt_asm+nt_prd),y_da(0:nt_asm+nt_prd),z_da(0:nt_asm+nt_prd)
  real(8) :: x_da_m(0:nt_asm+nt_prd,nens),y_da_m(0:nt_asm+nt_prd,nens),z_da_m(0:nt_asm+nt_prd,nens)
  real(8) :: x_obs(0:nt_asm/obs_interval)
  !---Output control
  integer, parameter :: output_interval=50
  character(len=7) :: obs_chr(0:nt_asm)
  !---Matrix
  real(8) :: M(3,3)  !state transient matrix
  real(8) :: Pf(3,3) !element 1:x, element 2:y, element 3:z
  real(8) :: Pa(3,3) !element 1:x, element 2:y, element 3:z
  real(8) :: R(1,1)  !element 1:x, element 2:y, element 3:z
  real(8) :: Kg(3,1) !element 1:x, element 2:y, element 3:z
  real(8) :: H(1,3)  !element 1:x, element 2:y, element 3:z
  !---Working variable
  integer :: it,iens
  real(8) :: x_innov
  real(8) :: x_prtb(nens),y_prtb(nens),z_prtb(nens)
  real(8) :: Ptmp(3,3)
  real(8) :: noise1,noise2,gnoise

  !*****Initialization of random number generator
  call random_seed()

  !*****Section 1: Initial value setting
  x_t(0)=10.0d0
  y_t(0)=14.0d0
  z_t(0)=24.0d0
  x_s(0)=11.0d0
  y_s(0)=13.0d0
  z_s(0)=25.0d0
  Pf(1,1)=1.0d0; Pf(1,2)=0.0d0; Pf(1,3)=0.0d0
  Pf(2,1)=0.0d0; Pf(2,2)=1.0d0; Pf(2,3)=0.0d0
  Pf(3,1)=0.0d0; Pf(3,2)=0.0d0; Pf(3,3)=1.0d0
  Pa=Pf
  R(1,1)=0.1d0
 
  !*****Section 2: True Field and Observations
  do it=1,nt_asm+nt_prd
    !---forward time step
    x_t(it) = x_t(it-1)                                                   &
    &         - dt * parm1*(x_t(it-1) - y_t(it-1))
    y_t(it) = y_t(it-1)                                                   &
    &       + dt * (-y_t(it-1) - x_t(it-1) * z_t(it-1) + parm2 * x_t(it-1))
    z_t(it) = z_t(it-1)                                                   &
    &       + dt * (x_t(it-1) * y_t(it-1) - parm3 * z_t(it-1))
    !---Observation
    if ((mod(it,obs_interval) == 0) .and. (it <= nt_asm)) then
      ! Generate Gaussian Noise from uniform random number (Box-Muller)
      call random_number(noise1)
      call random_number(noise2)
      gnoise=sqrt(R(1,1))*sqrt(-2.0d0*log(1.0d0-noise1))*cos(2.0d0*pi*noise2)      ! Generate observation by adding Gaussian noise to true value
      x_obs(it/obs_interval)=x_t(it)+gnoise
    end if
  end do

  !*****Section 3: Simulation without DA
  do it=1,nt_asm+nt_prd
    x_s(it) = x_s(it-1)                                                   &
    &         - dt * parm1*(x_s(it-1) - y_s(it-1))
    y_s(it) = y_s(it-1)                                                   &
    &       + dt * (-y_s(it-1) - x_s(it-1) * z_s(it-1) + parm2 * x_s(it-1))
    z_s(it) = z_s(it-1)                                                   &
    &       + dt * (x_s(it-1) * y_s(it-1) - parm3 * z_s(it-1))
  end do

  !*****Section 4: Data Assimilation
  do iens=1,nens
    ! Generate Gaussian Noise (gnoise) from uniform random number 
    call random_number(noise1)
    call random_number(noise2)
    gnoise=sqrt(R(1,1))*sqrt(-2.0d0*log(1.0d0-noise1))*cos(2.0d0*pi*noise2)
    x_da_m(0,iens)=x_s(0)+gnoise
    ! Generate Gaussian Noise (gnoise) from uniform random number 
    call random_number(noise1)
    call random_number(noise2)
    gnoise=sqrt(R(1,1))*sqrt(-2.0d0*log(1.0d0-noise1))*cos(2.0d0*pi*noise2)
    y_da_m(0,iens)=y_s(0)+gnoise
    ! Generate Gaussian Noise (gnoise) from uniform random number 
    call random_number(noise1)
    call random_number(noise2)
    gnoise=sqrt(R(1,1))*sqrt(-2.0d0*log(1.0d0-noise1))*cos(2.0d0*pi*noise2)
    z_da_m(0,iens)=z_s(0)+gnoise
  end do
  x_da(0)=sum(x_da_m(0,1:nens))/nens
  y_da(0)=sum(y_da_m(0,1:nens))/nens
  z_da(0)=sum(z_da_m(0,1:nens))/nens

  do it=1,nt_asm
    !---time integration
    do iens=1,nens
      x_da_m(it,iens) = x_da_m(it-1,iens)                                &
      & - dt * parm1*(x_da_m(it-1,iens) - y_da_m(it-1,iens))
      y_da_m(it,iens) = y_da_m(it-1,iens)                                &
      & + dt * (-y_da_m(it-1,iens) - x_da_m(it-1,iens)*z_da_m(it-1,iens) + parm2*x_da_m(it-1,iens))
      z_da_m(it,iens) = z_da_m(it-1,iens)                                &
      & + dt*(x_da_m(it-1,iens)*y_da_m(it-1,iens) - parm3*z_da_m(it-1,iens))
    end do
    ! Section 4-2-1 Obtain Pf
    if(mod(it,obs_interval) == 0) then
      x_da(it)=sum(x_da_m(it,1:nens))/nens
      y_da(it)=sum(y_da_m(it,1:nens))/nens
      z_da(it)=sum(z_da_m(it,1:nens))/nens
      Pf=0.0
      do iens=1,nens
        x_prtb(iens)=x_da_m(it,iens)-x_da(it)
        y_prtb(iens)=y_da_m(it,iens)-y_da(it)
        z_prtb(iens)=z_da_m(it,iens)-z_da(it)
        Pf(1,1)=Pf(1,1)+x_prtb(iens)**2/(nens-1)
        Pf(1,2)=Pf(1,2)+x_prtb(iens)*y_prtb(iens)/(nens-1)
        Pf(1,3)=Pf(1,3)+x_prtb(iens)*z_prtb(iens)/(nens-1)
        Pf(2,1)=Pf(2,1)+y_prtb(iens)*x_prtb(iens)/(nens-1)
        Pf(2,2)=Pf(2,2)+y_prtb(iens)**2/(nens-1)
        Pf(2,3)=Pf(2,3)+y_prtb(iens)*z_prtb(iens)/(nens-1)
        Pf(3,1)=Pf(3,1)+z_prtb(iens)*x_prtb(iens)/(nens-1)
        Pf(3,2)=Pf(3,2)+z_prtb(iens)*y_prtb(iens)/(nens-1)
        Pf(3,3)=Pf(3,3)+z_prtb(iens)**2/(nens-1)
      end do
      ! Section 4-2-3: Kalman gain: Weighting of model result and obs.
      ! (Note) Obsevation only in x --> component 1(x) only
      !        In this case, inverse matrix --> scalar inverse
      Kg(1,1)=Pf(1,1)/(R(1,1)+Pf(1,1))
      Kg(2,1)=Pf(2,1)/(R(1,1)+Pf(1,1))
      Kg(3,1)=Pf(3,1)/(R(1,1)+Pf(1,1))
      ! Section 4-2-4: calculate innovation
      ! Section 4-2-3: calculate innovation and correction
      do iens=1,nens
        ! Generate Gaussian Noise (gnoise) from uniform random number 
        call random_number(noise1)
        call random_number(noise2)
        gnoise=sqrt(R(1,1))*sqrt(-2.0*log(1.0 - noise1))*cos(2 * pi * noise2)
        x_innov=x_obs(it/obs_interval)+gnoise-x_da_m(it,iens)
        x_da_m(it,iens)=x_da_m(it,iens)+Kg(1,1)*x_innov
        y_da_m(it,iens)=y_da_m(it,iens)+Kg(2,1)*x_innov
        z_da_m(it,iens)=z_da_m(it,iens)+Kg(3,1)*x_innov
      end do
      ! Section 4-2-5: analysis error covariance matrix
      Pa=Pf-matmul(matmul(Kg,H),Pf)
      Pf=Pa
      write(*,'(A,F8.2,A,9F8.3)') "time=",dt*it, ", Pa=", Pa
    end if
    x_da(it)=sum(x_da_m(it,1:nens))/nens
    y_da(it)=sum(y_da_m(it,1:nens))/nens
    z_da(it)=sum(z_da_m(it,1:nens))/nens
  end do

  !*****Section 5: Prediction after Data Assimilation
  do it=nt_asm+1,nt_asm+nt_prd
    x_da(it) = x_da(it-1)                                                 &
    &         - dt * parm1*(x_da(it-1) - y_da(it-1))
    y_da(it) = y_da(it-1)                                                 &
    &       + dt * (-y_da(it-1) - x_da(it-1)*z_da(it-1) + parm2*x_da(it-1))
    z_da(it) = z_da(it-1)                                                   &
    &       + dt * (x_da(it-1) * y_da(it-1) - parm3 * z_da(it-1))
  end do

  !*****Output on screen
  !---Preparation output: conversion real --> character
  obs_chr(0:nt_asm)="No obs"
  do it=1,nt_asm
    if (mod(it,obs_interval) == 0) then
      write(obs_chr(it), '(F7.3)') x_obs(it/obs_interval)
    end if
  end do
  !---Output Main
  write(*,*) "#######################################"
  write(*,*) "###    Identical Twin Experiment    ###"
  write(*,*) "###    Ensemble Kalman Filter       ###"
  write(*,*) "#######################################"
  write(*,*) 
  write(*,'(A,F7.2,A,F7.2)') "Assmilation Period: t=",        &
  &                          0.0,"-",dt*nt_asm
  write(*,'(A,F7.2,A,F7.2)') "Prediction Period:  t=",         &
  &                          dt*nt_asm,"-", dt*(nt_asm+nt_prd)
  write(*,*) 
  write(*,*) "*******Assimilation Period: x ********"
  write(*,*) " [Time]   [True]  [No Assim]  [Assim]  [Observation]"
  do it=0,nt_asm
    if (mod(it,output_interval) == 0) then
      write(*,'(F7.2,3F10.3,4X,A)') dt*it, x_t(it), x_s(it),x_da(it),obs_chr(it)
    endif
  end do
  write(*,*) "*******Prediction Period: x ********"
  do it=nt_asm+1,nt_asm+nt_prd
    if (mod(it,output_interval) == 0) then
      write(*,'(F7.2,3F10.3)') dt*it, x_t(it), x_s(it),x_da(it)
    endif
  end do
  write(*,*)
  write(*,*) "*******Assimilation Period: y ********"
  write(*,*) " [Time]   [True]  [No Assim]  [Assim]"
  do it=0,nt_asm
    if (mod(it,output_interval) == 0) then
      write(*,'(F7.2,3F10.3)') dt*it, y_t(it), y_s(it), y_da(it)
    endif
  end do
  write(*,*) "*******Prediction Period: y ********"
  do it=nt_asm+1,nt_asm+nt_prd
    if (mod(it,output_interval) == 0) then
      write(*,'(F7.2,3F10.3)') dt*it, y_t(it), y_s(it), y_da(it)
    endif
  end do
  write(*,*)
  write(*,*) "*******Assimilation Period: z ********"
  write(*,*) " [Time]   [True]  [No Assim]  [Assim]"
  do it=0,nt_asm
    if (mod(it,output_interval) == 0) then
      write(*,'(F7.2,3F10.3)') dt*it, z_t(it), z_s(it), z_da(it)
    endif
  end do
  write(*,*) "*******Prediction Period: z ********"
  do it=nt_asm+1,nt_asm+nt_prd
    if (mod(it,output_interval) == 0) then
      write(*,'(F7.2,3F10.3)') dt*it, z_t(it), z_s(it), z_da(it)
    endif
  end do
  
  end program Lorenz63_EnKF
