!***********************
! Lorenz 1963 (Extended KF)
! 2012. 8.29 K. Ito
!***********************

  program Lorenz63_KF
  implicit none
  
  !*****
  !---Parameters
  integer, parameter :: nt_asm=1000, nt_prd=1000
  integer, parameter :: obs_interval=100
  real(8), parameter :: parm1=10.0d0, parm2=28.0d0, parm3=2.667d0, dt=1.0d-3
  real(8), parameter :: pi=3.14159265358979
  !---Physical variable
  real(8) :: x_t(0:nt_asm+nt_prd), y_t(0:nt_asm+nt_prd), z_t(0:nt_asm+nt_prd) 
  real(8) :: x_s(0:nt_asm+nt_prd), y_s(0:nt_asm+nt_prd), z_s(0:nt_asm+nt_prd) 
  real(8) :: x_da(0:nt_asm+nt_prd),y_da(0:nt_asm+nt_prd),z_da(0:nt_asm+nt_prd)
  real(8) :: x_obs(0:nt_asm/obs_interval)
  !---Output control
  integer, parameter :: output_interval=50
  character(len=7) :: obs_chr(0:nt_asm)
  !---Matrix
  real(8) :: M(3,3)  !state transient matrix
  real(8) :: Pf(3,3) !element 1:x, element 2:y, element 3:z
  real(8) :: Pa(3,3) !element 1:x, element 2:y, element 3:z
  real(8) :: R(1,1)  !element 1:x, element 2:y, element 3:z
  real(8) :: Kg(3,1) !element 1:x, element 2:y, element 3:z
  real(8) :: H(1,3)  !element 1:x, element 2:y, element 3:z
  !---Working variable
  integer :: it
  real(8) :: x_innov
  real(8) :: Ptmp(3,3), MatTemp(3,3)
  real(8) :: noise1,noise2,gnoise

  !*****Initialization of random number generator
  call random_seed()

  !*****Section 1: Initial value setting
  x_t(0)=10.0d0
  y_t(0)=14.0d0
  z_t(0)=24.0d0
  x_s(0)=11.0d0
  y_s(0)=13.0d0
  z_s(0)=25.0d0
  Pf(1,1)=1.0d0; Pf(1,2)=0.0d0; Pf(1,3)=0.0d0
  Pf(2,1)=0.0d0; Pf(2,2)=1.0d0; Pf(2,3)=0.0d0
  Pf(3,1)=0.0d0; Pf(3,2)=0.0d0; Pf(3,3)=1.0d0
  Pa=Pf
  R(1,1)=0.1d0
  H(1,1)=1.0d0; H(1,2)=0.0d0; H(1,3)=0.0d0
  
  !*****Section 2: True Field and Observations
  do it=1,nt_asm+nt_prd
    !---forward time step
    x_t(it) = x_t(it-1)                                                   &
    &         - dt * parm1*(x_t(it-1) - y_t(it-1))
    y_t(it) = y_t(it-1)                                                   &
    &       + dt * (-y_t(it-1) - x_t(it-1) * z_t(it-1) + parm2 * x_t(it-1))
    z_t(it) = z_t(it-1)                                                   &
    &       + dt * (x_t(it-1) * y_t(it-1) - parm3 * z_t(it-1))
    !---Observation
    if ((mod(it,obs_interval) == 0) .and. (it <= nt_asm)) then
      ! Generate Gaussian Noise from uniform random number (Box-Muller)
      call random_number(noise1)
      call random_number(noise2)
      gnoise=sqrt(R(1,1))*sqrt(-2.0d0*log(1.0d0-noise1))*cos(2.0d0*pi*noise2)
      ! Generate observation by adding Gaussian noise to true value
      x_obs(it/obs_interval)=x_t(it)+gnoise
    end if
  end do

  !*****Section 3: Simulation without DA
  do it=1,nt_asm+nt_prd
    x_s(it) = x_s(it-1)                                                   &
    &         - dt * parm1*(x_s(it-1) - y_s(it-1))
    y_s(it) = y_s(it-1)                                                   &
    &       + dt * (-y_s(it-1) - x_s(it-1) * z_s(it-1) + parm2 * x_s(it-1))
    z_s(it) = z_s(it-1)                                                   &
    &       + dt * (x_s(it-1) * y_s(it-1) - parm3 * z_s(it-1))
  end do

  !*****Section 4: Data Assimilation
  x_da(0)=x_s(0)
  y_da(0)=y_s(0)
  z_da(0)=z_s(0)
  do it=1,nt_asm
    !---time integration
    x_da(it) = x_da(it-1)                                                 &
    &         - dt * parm1*(x_da(it-1) - y_da(it-1))
    y_da(it) = y_da(it-1)                                                 &
    &       + dt * (-y_da(it-1) - x_da(it-1)*z_da(it-1) + parm2*x_da(it-1))
    z_da(it) = z_da(it-1)                                                   &
    &       + dt * (x_da(it-1) * y_da(it-1) - parm3 * z_da(it-1))
    ! State Transient Matrix
    M(1,1)=1.0d0-dt*parm1;       M(1,2)=dt*parm1;    M(1,3)=0.0d0
    M(2,1)=dt*(-z_da(it)+parm2); M(2,2)=-dt+1.0d0;   M(2,3)=-dt*x_da(it)
    M(3,1)=dt*y_da(it);          M(3,2)=dt*x_da(it); M(3,3)=1.0d0-dt*parm3
    ! Section 4-2-2 Lyapunov equation: Obtain Pf
    Ptmp=transpose(M)    !Transpose matrix
    Ptmp=matmul(Pf,Ptmp) !Matrix maltiplication
    Pf=matmul(M,Ptmp)    !Matrix maltiplication
!    write(*,'(A,F10.2,A,9F8.3)') "time=",dt*it, ", Pf=", Pf
    if(mod(it,obs_interval) == 0) then
      ! Section 4-2-3: Kalman gain: Weighting of model result and obs.
      Kg(1,1)=Pf(1,1)/(R(1,1)+Pf(1,1))
      Kg(2,1)=Pf(2,1)/(R(1,1)+Pf(1,1))
      Kg(3,1)=Pf(3,1)/(R(1,1)+Pf(1,1))
      ! Section 4-2-4: calculate innovation
      x_innov=x_obs(it/obs_interval)-x_da(it)
      x_da(it)=x_da(it)+Kg(1,1)*x_innov
      y_da(it)=y_da(it)+Kg(2,1)*x_innov
      z_da(it)=z_da(it)+Kg(3,1)*x_innov
      ! Section 4-2-5: analysis error covariance matrix
      Pa=Pf-matmul(matmul(Kg,H),Pf)
      Pf=Pa
      write(*,'(A,F8.2,A,9F9.3)') "time=",dt*it, ", Pa=", Pa
    end if
  end do

  !*****Section 5: Prediction after Data Assimilation
  do it=nt_asm+1,nt_asm+nt_prd
    x_da(it) = x_da(it-1)                                                 &
    &         - dt * parm1*(x_da(it-1) - y_da(it-1))
    y_da(it) = y_da(it-1)                                                 &
    &       + dt * (-y_da(it-1) - x_da(it-1)*z_da(it-1) + parm2*x_da(it-1))
    z_da(it) = z_da(it-1)                                                   &
    &       + dt * (x_da(it-1) * y_da(it-1) - parm3 * z_da(it-1))
  end do

  !*****Output on screen
  !---Preparation output: conversion real --> character
  obs_chr(0:nt_asm)="No obs"
  do it=1,nt_asm
    if (mod(it,obs_interval) == 0) then
      write(obs_chr(it), '(F7.3)') x_obs(it/obs_interval)
    end if
  end do
  !---Output Main
  write(*,*) "#######################################"
  write(*,*) "###    Identical Twin Experiment    ###"
  write(*,*) "###    Lorenz63: Kalman Filter      ###"
  write(*,*) "#######################################"
  write(*,*) 
  write(*,'(A,F7.2,A,F7.2)') "Assmilation Period: t=",        &
  &                          0.0,"-",dt*nt_asm
  write(*,'(A,F7.2,A,F7.2)') "Prediction Period:  t=",         &
  &                          dt*nt_asm,"-", dt*(nt_asm+nt_prd)
  write(*,*) 
  write(*,*) "*******Assimilation Period: x ********"
  write(*,*) " [Time]   [True]  [No Assim]  [Assim]  [Observation]"
  do it=0,nt_asm
    if (mod(it,output_interval) == 0) then
      write(*,'(F7.2,3F10.3,4X,A)') dt*it, x_t(it), x_s(it),x_da(it),obs_chr(it)
    endif
  end do
  write(*,*) "*******Prediction Period: x ********"
  do it=nt_asm+1,nt_asm+nt_prd
    if (mod(it,output_interval) == 0) then
      write(*,'(F7.2,3F10.3)') dt*it, x_t(it), x_s(it),x_da(it)
    endif
  end do
  write(*,*)
  write(*,*) "*******Assimilation Period: y ********"
  write(*,*) " [Time]   [True]  [No Assim]  [Assim]"
  do it=0,nt_asm
    if (mod(it,output_interval) == 0) then
      write(*,'(F7.2,3F10.3)') dt*it, y_t(it), y_s(it), y_da(it)
    endif
  end do
  write(*,*) "*******Prediction Period: y ********"
  do it=nt_asm+1,nt_asm+nt_prd
    if (mod(it,output_interval) == 0) then
      write(*,'(F7.2,3F10.3)') dt*it, y_t(it), y_s(it), y_da(it)
    endif
  end do
  write(*,*)
  write(*,*) "*******Assimilation Period: z ********"
  write(*,*) " [Time]   [True]  [No Assim]  [Assim]"
  do it=0,nt_asm
    if (mod(it,output_interval) == 0) then
      write(*,'(F7.2,3F10.3)') dt*it, z_t(it), z_s(it), z_da(it)
    endif
  end do
  write(*,*) "*******Prediction Period: z ********"
  do it=nt_asm+1,nt_asm+nt_prd
    if (mod(it,output_interval) == 0) then
      write(*,'(F7.2,3F10.3)') dt*it, z_t(it), z_s(it), z_da(it)
    endif
  end do
  
  end program Lorenz63_KF
